/* PyAuthD_PPP.c
   -------------

   Implements the password checking logic of pppd. It plugs into the standard
   pppd chap_verify and pap_verify architecture, and also allows you to
   specify an IP-address which is to be used by the remote end.

   Copyright (C) 2005, Heiko Wundram.
   Released under a BSD-style license, see LICENSE for details.

$LastChangedBy: modelnine $
$LastChangedRevision: 75 $
$LastChangedDate: 2005-10-24 01:51:59 +0200 (pon, 24 paź 2005) $
*/

#include <malloc.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/un.h>
#include <syslog.h>
#include <unistd.h>

/* Load all proper specifications for ppp. */

#include "ppp/pppd.h"
#include "ppp/chap-new.h"

/* Load specifications for PyAuthD and PyAuthD_ppp. */

#include "PyAuthD_client.h"
#include "PyAuthD_PPP.h"

/* Globals. */

char pppd_version[] = VERSION;
int sock;

void plugin_init()
{
  pap_check_hook = &PyAuthD_secret_check;
  pap_auth_hook = &PyAuthD_pap_auth;

  chap_check_hook = &PyAuthD_secret_check;
  chap_verify_hook = &PyAuthD_chap_auth;

  null_auth_hook = &PyAuthD_null_auth;

  ip_choose_hook = &PyAuthD_choose_ip;
  allowed_address_hook = &PyAuthD_allow_ip;

  idle_time_hook = &PyAuthD_check_idle;
  holdoff_hook = &PyAuthD_check_holdoff;

  add_notifier(&exitnotify,&plugin_close,NULL);
  add_notifier(&pidchange,&PyAuthD_notify,(void*)AUTHD_PPP_SIGNAL_PIDCHANGE);
  add_notifier(&phasechange,&PyAuthD_notify,
	       (void*)AUTHD_PPP_SIGNAL_PHASECHANGE);
  add_notifier(&sigreceived,&PyAuthD_notify,
	       (void*)AUTHD_PPP_SIGNAL_SIGRECEIVED);
  add_notifier(&ip_up_notifier,&PyAuthD_notify,
	       (void*)AUTHD_PPP_SIGNAL_IPUP);
  add_notifier(&ip_down_notifier,&PyAuthD_notify,
	       (void*)AUTHD_PPP_SIGNAL_IPDOWN);
  add_notifier(&auth_up_notifier,&PyAuthD_notify,
	       (void*)AUTHD_PPP_SIGNAL_AUTHUP);
  add_notifier(&link_down_notifier,&PyAuthD_notify,
	       (void*)AUTHD_PPP_SIGNAL_LINKDOWN);

  sock = openPyAuthD(AUTHD_PPP);
}

void plugin_close(void *opaque, int arg)
{
  int comm = AUTHD_PPP_SIGNAL_EXIT;

  if( sock == -2 )
    return;
  if( sock == -1 )
    goto SOCKET_ERROR;

  SEND_VALUE(comm);

  closePyAuthD(sock);
  sock = -1;

  return;

 SEND_ERROR:
 SOCKET_ERROR:
  sock = -1;
  syslog(LOG_AUTHPRIV|LOG_CRIT,"Error while talking to PyAuthD.");
}

static int PyAuthD_secret_check()
{
  if( sock == -2 )
    return -1;
  return 1;
}

static int PyAuthD_pap_auth(char *user, char *passwd, char **msgp,
			    struct wordlist **paddrs, struct wordlist **popts)
{
  int comm = AUTHD_PPP_GET_SECRET, secretlen, rv;
  char *secret;

  if( sock == -2 )
    return -1;
  if( sock == -1 )
    goto SOCKET_ERROR;

  SEND_VALUE(comm);
  SEND_DATA(user,secretlen);
  secretlen = -1;
  SEND_VALUE(secretlen);
  RECV_VALUE(secretlen);
  if( secretlen != -1 ) {
    RECV_DATA(secret,secretlen);
    rv = strcmp(passwd,secret) == 0;
    FREE_DATA(secret,secretlen);
  } else
    rv = 0;

  return rv;

 SEND_ERROR:
 MEMORY_ERROR:
 RECV_ERROR:
 SOCKET_ERROR:
  sock = -1;
  syslog(LOG_AUTHPRIV|LOG_CRIT,"Error while talking to PyAuthD.");
  return 0;
}

static int PyAuthD_chap_auth(char *user, char *ourname, int id,
			     struct chap_digest_type *digest,
			     unsigned char *challenge, unsigned char *response,
			     unsigned char *message, int message_space)
{
  int comm = AUTHD_PPP_GET_SECRET, secretlen, rv;
  char *secret;

  if( sock == -2 )
    return CHAP_FAILURE;
  if( sock == -1 )
    goto SOCKET_ERROR;

  SEND_VALUE(comm);
  SEND_DATA(user,secretlen);
  SEND_DATA(ourname,secretlen);
  RECV_VALUE(secretlen);
  if( secretlen != -1 ) {
    RECV_DATA(secret,secretlen);
    rv = digest->verify_response(id,user,secret,secretlen,challenge,response,
				 message,message_space);
    FREE_DATA(secret,secretlen);
  } else
    rv = 0;

  return rv;

 SEND_ERROR:
 MEMORY_ERROR:
 RECV_ERROR:
 SOCKET_ERROR:
  sock = -1;
  syslog(LOG_AUTHPRIV|LOG_CRIT,"Error while talking to PyAuthD.");
  return CHAP_FAILURE;
}

static int PyAuthD_null_auth(struct wordlist **paddrs, struct wordlist **popts)
{
  int comm = AUTHD_PPP_GET_SECRET, rv;

  if( sock == -2 )
    return -1;
  if( sock == -1 )
    goto SOCKET_ERROR;

  SEND_VALUE(comm);
  rv = -1;
  SEND_VALUE(rv);
  RECV_VALUE(rv);

  return rv;

 SEND_ERROR:
 RECV_ERROR:
 SOCKET_ERROR:
  sock = -1;
  syslog(LOG_AUTHPRIV|LOG_CRIT,"Error while talking to PyAuthD.");
  return 0;
}

static void PyAuthD_choose_ip(u_int32_t *addrp)
{
  int comm = AUTHD_PPP_CHOOSE_IP;

  if( sock == -2 )
    return;
  if( sock == -1 )
    goto SOCKET_ERROR;

  SEND_VALUE(comm);
  SEND_VALUE((*addrp));
  RECV_VALUE((*addrp));

  return;

 SEND_ERROR:
 RECV_ERROR:
 SOCKET_ERROR:
  sock = -1;
  syslog(LOG_AUTHPRIV|LOG_CRIT,"Error while talking to PyAuthD.");
}

static int PyAuthD_allow_ip(u_int32_t addr)
{
  int comm = AUTHD_PPP_CHECK_IP, rv;

  if( sock == -2 )
    return -1;
  if( sock == -1 )
    goto SOCKET_ERROR;

  SEND_VALUE(comm);
  SEND_VALUE(addr);
  RECV_VALUE(rv);

  return rv;

 SEND_ERROR:
 RECV_ERROR:
 SOCKET_ERROR:
  sock = -1;
  syslog(LOG_AUTHPRIV|LOG_CRIT,"Error while talking to PyAuthD.");
  return 0;
}

static int PyAuthD_check_idle(struct ppp_idle *idlep)
{
  int comm = AUTHD_PPP_CHECK_IDLE, rv;

  if( sock == -2 )
    return 0;
  if( sock == -1 )
    goto SOCKET_ERROR;

  SEND_VALUE(comm);
  if( idlep ) {
    SEND_VALUE(idlep->xmit_idle);
    SEND_VALUE(idlep->recv_idle);
  } else {
    rv = -1;
    SEND_VALUE(rv);
  }
  RECV_VALUE(rv);

  return rv;

 SEND_ERROR:
 RECV_ERROR:
 SOCKET_ERROR:
  sock = -1;
  syslog(LOG_AUTHPRIV|LOG_CRIT,"Error while talking to PyAuthD.");
  return 0;
}

static int PyAuthD_check_holdoff()
{
  int comm = AUTHD_PPP_CHECK_HOLDOFF, rv;

  if( sock == -2 )
    return 0;
  if( sock == -1 )
    goto SOCKET_ERROR;

  SEND_VALUE(comm);
  RECV_VALUE(rv);

  return rv;

 SEND_ERROR:
 RECV_ERROR:
 SOCKET_ERROR:
  sock = -1;
  syslog(LOG_AUTHPRIV|LOG_CRIT,"Error while talking to PyAuthD.");
  return 0;
}

void PyAuthD_notify(void *opaque, int arg)
{
  int comm = (int)opaque, rv;

  if( sock == -2 )
    return;
  if( sock == -1 )
    goto SOCKET_ERROR;

  SEND_VALUE(comm);
  SEND_VALUE(arg);
  RECV_VALUE(rv);

  if( rv == AUTHD_PPP_CLOSE_ERROR ) {
    closePyAuthD(sock);
    sock = -1;
  } else if( rv == AUTHD_PPP_CLOSE_FINISH ) {
    closePyAuthD(sock);
    sock = -2;
  }

  return;

 SEND_ERROR:
 RECV_ERROR:
 SOCKET_ERROR:
  sock = -1;
  syslog(LOG_AUTHPRIV|LOG_CRIT,"Error while talking to PyAuthD.");
}
